class baseML(object):
    def __init__(self):
        self.model = None
        pass

    def train(self, data):
        pass

    def predict(self, x):
        pass

    def save(self, path):
        pass
